# This file is part of rddl2tf.

# rddl2tf is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# rddl2tf is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with rddl2tf. If not, see <http://www.gnu.org/licenses/>.


import rddlgym

from pyrddl.expr import Expression
from pyrddl import utils

from rddl2tf.compiler import Compiler
from rddl2tf.fluent import TensorFluent
from rddl2tf.fluentshape import TensorFluentShape

import numpy as np
import tensorflow as tf

import unittest


class TestCompiler(unittest.TestCase):

    def setUp(self):
        self.rddl1 = rddlgym.make('Reservoir-8', mode=rddlgym.AST)
        self.rddl2 = rddlgym.make('Mars_Rover', mode=rddlgym.AST)
        self.rddl3 = rddlgym.make('HVAC-v1', mode=rddlgym.AST)
        self.rddl4 = rddlgym.make('CrossingTraffic-10', mode=rddlgym.AST)
        self.rddl5 = rddlgym.make('GameOfLife-10', mode=rddlgym.AST)
        self.rddl6 = rddlgym.make('CarParking-v1', mode=rddlgym.AST)
        self.rddl7 = rddlgym.make('Navigation-v3', mode=rddlgym.AST)
        self.compiler1 = Compiler(self.rddl1)
        self.compiler2 = Compiler(self.rddl2)
        self.compiler3 = Compiler(self.rddl3)
        self.compiler4 = Compiler(self.rddl4)
        self.compiler5 = Compiler(self.rddl5)
        self.compiler6 = Compiler(self.rddl6)
        self.compiler7 = Compiler(self.rddl7)

    def test_compile_state_action_constraints(self):
        batch_size = 1000
        compilers = [self.compiler4, self.compiler5]
        expected_preconds = [(12, [True] + [False] * 11), (1, [False])]
        for compiler, expected in zip(compilers, expected_preconds):
            compiler.batch_mode_on()
            state = compiler.compile_initial_state(batch_size)
            action = compiler.compile_default_action(batch_size)
            constraints = compiler.compile_state_action_constraints(state, action)
            self.assertIsInstance(constraints, list)
            self.assertEqual(len(constraints), expected[0])
            for c, batch_mode in zip(constraints, expected[1]):
                self.assertIsInstance(c, TensorFluent)
                self.assertEqual(c.dtype, tf.bool)
                if batch_mode:
                    self.assertEqual(c.shape.batch_size, batch_size)
                else:
                    self.assertEqual(c.shape.batch_size, 1)
                self.assertEqual(c.shape.batch, batch_mode)
                self.assertTupleEqual(c.shape.fluent_shape, ())

    def test_compile_action_preconditions(self):
        batch_size = 1000
        compilers = [self.compiler1, self.compiler2]
        expected_preconds = [2, 1]
        for compiler, expected in zip(compilers, expected_preconds):
            compiler.batch_mode_on()
            state = compiler.compile_initial_state(batch_size)
            action = compiler.compile_default_action(batch_size)
            preconds = compiler.compile_action_preconditions(state, action)
            self.assertIsInstance(preconds, list)
            self.assertEqual(len(preconds), expected)
            for p in preconds:
                self.assertIsInstance(p, TensorFluent)
                self.assertEqual(p.dtype, tf.bool)
                self.assertEqual(p.shape.batch_size, batch_size)
                self.assertTrue(p.shape.batch)
                self.assertTupleEqual(p.shape.fluent_shape, ())

    def test_compile_state_invariants(self):
        batch_size = 1000
        compilers = [self.compiler1, self.compiler2]
        expected_invariants = [2, 0]
        for compiler, expected in zip(compilers, expected_invariants):
            compiler.batch_mode_on()
            state = compiler.compile_initial_state(batch_size)
            invariants = compiler.compile_state_invariants(state)
            self.assertIsInstance(invariants, list)
            self.assertEqual(len(invariants), expected)
            for p in invariants:
                self.assertIsInstance(p, TensorFluent)
                self.assertEqual(p.dtype, tf.bool)
                self.assertTupleEqual(p.shape.fluent_shape, ())

    def test_compile_action_preconditions_checking(self):
        batch_size = 1000
        compilers = [self.compiler1, self.compiler2]
        for compiler in compilers:
            compiler.batch_mode_on()
            state = compiler.compile_initial_state(batch_size)
            action = compiler.compile_default_action(batch_size)
            checking = compiler.compile_action_preconditions_checking(state, action)
            self.assertIsInstance(checking, tf.Tensor)
            self.assertEqual(checking.dtype, tf.bool)
            self.assertListEqual(checking.shape.as_list(), [batch_size])

    def test_compile_action_lower_bound_constraints(self):
        batch_size = 1000
        compilers = [self.compiler1, self.compiler3]
        expected = [[('outflow/1', [], tf.int32)], [('AIR/1', [], tf.int32)]]

        for compiler, expected_bounds in zip(compilers, expected):
            compiler.batch_mode_on()
            initial_state = compiler.compile_initial_state(batch_size)
            default_action_fluents = compiler.compile_default_action(batch_size)
            bounds = compiler.compile_action_bound_constraints(initial_state)
            self.assertIsInstance(bounds, dict)

            self.assertEqual(len(bounds), len(expected_bounds))
            for fluent_name, shape, dtype in expected_bounds:
                self.assertIn(fluent_name, bounds)
                self.assertIsInstance(bounds[fluent_name], tuple)
                self.assertEqual(len(bounds[fluent_name]), 2)
                lower, _ = bounds[fluent_name]
                self.assertIsInstance(lower, TensorFluent)
                self.assertListEqual(lower.shape.as_list(), shape)
                self.assertEqual(lower.dtype, dtype)

    def test_compile_action_lower_bound_constraints(self):
        batch_size = 1000
        compilers = [self.compiler1, self.compiler3]
        expected = [[('outflow/1', [], tf.int32)], [('AIR/1', [], tf.int32)]]

        for compiler, expected_bounds in zip(compilers, expected):
            compiler.batch_mode_on()
            initial_state = compiler.compile_initial_state(batch_size)
            default_action_fluents = compiler.compile_default_action(batch_size)
            bounds = compiler.compile_action_bound_constraints(initial_state)
            self.assertIsInstance(bounds, dict)

            self.assertEqual(len(bounds), len(expected_bounds))
            for fluent_name, shape, dtype in expected_bounds:
                self.assertIn(fluent_name, bounds)
                self.assertIsInstance(bounds[fluent_name], tuple)
                self.assertEqual(len(bounds[fluent_name]), 2)
                lower, _ = bounds[fluent_name]
                self.assertIsInstance(lower, TensorFluent)
                self.assertListEqual(lower.shape.as_list(), shape)
                self.assertEqual(lower.dtype, dtype)

    def test_compile_action_upper_bound_constraints(self):
        batch_size = 1000
        compilers = [self.compiler1, self.compiler3]
        expected = [[('outflow/1', [batch_size, 8], tf.float32)], [('AIR/1', [3], tf.float32)]]

        for compiler, expected_bounds in zip(compilers, expected):
            compiler.batch_mode_on()
            initial_state = compiler.compile_initial_state(batch_size)
            default_action_fluents = compiler.compile_default_action(batch_size)
            bounds = compiler.compile_action_bound_constraints(initial_state)
            self.assertIsInstance(bounds, dict)

            self.assertEqual(len(bounds), len(expected_bounds))
            for fluent_name, shape, dtype in expected_bounds:
                self.assertIn(fluent_name, bounds)
                self.assertIsInstance(bounds[fluent_name], tuple)
                self.assertEqual(len(bounds[fluent_name]), 2)
                _, upper = bounds[fluent_name]
                self.assertIsInstance(upper, TensorFluent)
                self.assertEqual(upper.dtype, dtype)
                self.assertListEqual(upper.shape.as_list(), shape)

    def test_initialize_non_fluents(self):
        nf = dict(self.compiler1.compile_non_fluents())

        expected_non_fluents = {
            'MAX_RES_CAP/1': { 'shape': [8,], 'dtype': tf.float32 },
            'UPPER_BOUND/1': { 'shape': [8,], 'dtype': tf.float32 },
            'LOWER_BOUND/1': { 'shape': [8,], 'dtype': tf.float32 },
            'RAIN_SHAPE/1': { 'shape': [8,], 'dtype': tf.float32 },
            'RAIN_SCALE/1': { 'shape': [8,], 'dtype': tf.float32 },
            'DOWNSTREAM/2': { 'shape': [8,8], 'dtype': tf.bool },
            'SINK_RES/1': { 'shape': [8,], 'dtype': tf.bool },
            'MAX_WATER_EVAP_FRAC_PER_TIME_UNIT/0': { 'shape': [], 'dtype': tf.float32 },
            'LOW_PENALTY/1': { 'shape': [8,], 'dtype': tf.float32 },
            'HIGH_PENALTY/1': { 'shape': [8,], 'dtype': tf.float32 }
        }
        self.assertIsInstance(nf, dict)
        self.assertEqual(len(nf), len(expected_non_fluents))
        for name, fluent in nf.items():
            self.assertIn(name, expected_non_fluents)
            shape = expected_non_fluents[name]['shape']
            dtype = expected_non_fluents[name]['dtype']
            self.assertEqual(fluent.name, 'non_fluents/{}:0'.format(name.replace('/', '-')))
            self.assertIsInstance(fluent, TensorFluent)
            self.assertEqual(fluent.dtype, dtype)
            self.assertEqual(fluent.shape.as_list(), shape)

        expected_initializers = {
            'MAX_RES_CAP/1': [ 100.,  100.,  200.,  300.,  400.,  500.,  800., 1000.],
            'UPPER_BOUND/1': [ 80.,  80., 180., 280., 380., 480., 780., 980.],
            'LOWER_BOUND/1': [20., 20., 20., 20., 20., 20., 20., 20.],
            'RAIN_SHAPE/1': [1., 1., 1., 1., 1., 1., 1., 1.],
            'RAIN_SCALE/1': [ 5.,  3.,  9.,  7., 15., 13., 25., 30.],
            'DOWNSTREAM/2': [
                [False, False, False, False, False, True, False, False],
                [False, False, True, False, False, False, False, False],
                [False, False, False, False, True, False, False, False],
                [False, False, False, False, False, False, False, True],
                [False, False, False, False, False, False, True, False],
                [False, False, False, False, False, False, True, False],
                [False, False, False, False, False, False, False, True],
                [False, False, False, False, False, False, False, False]
            ],
            'SINK_RES/1': [False, False, False, False, False, False, False, True],
            'MAX_WATER_EVAP_FRAC_PER_TIME_UNIT/0': 0.05,
            'LOW_PENALTY/1': [-5., -5., -5., -5., -5., -5., -5., -5.],
            'HIGH_PENALTY/1': [-10., -10., -10., -10., -10., -10., -10., -10.]
        }
        with tf.Session(graph=self.compiler1.graph) as sess:
            for name, fluent in nf.items():
                value = sess.run(fluent.tensor)
                list1 = list(value.flatten())
                list2 = list(np.array(expected_initializers[name]).flatten())
                for v1, v2 in zip(list1, list2):
                    self.assertAlmostEqual(v1, v2)

    def test_initialize_initial_state_fluents(self):
        sf = dict(self.compiler1.compile_initial_state())

        expected_state_fluents = {
            'rlevel/1': { 'shape': [8,] , 'dtype': tf.float32 }
        }
        self.assertIsInstance(sf, dict)
        self.assertEqual(len(sf), len(expected_state_fluents))
        for name, fluent in sf.items():
            self.assertIn(name, expected_state_fluents)
            shape = expected_state_fluents[name]['shape']
            dtype = expected_state_fluents[name]['dtype']
            self.assertEqual(fluent.name, 'initial_state/{}:0'.format(name.replace('/', '-')))
            self.assertIsInstance(fluent, TensorFluent)
            self.assertEqual(fluent.dtype, dtype)
            self.assertEqual(fluent.shape.as_list(), shape)

        expected_initializers = {
            'rlevel/1': [75., 50., 50., 50., 50., 50., 50., 50.]
        }
        with tf.Session(graph=self.compiler1.graph) as sess:
            for name, fluent in sf.items():
                value = sess.run(fluent.tensor)
                list1 = list(value.flatten())
                list2 = list(np.array(expected_initializers[name]).flatten())
                for v1, v2 in zip(list1, list2):
                    self.assertAlmostEqual(v1, v2)

    def test_initialize_default_action_fluents(self):
        action_fluents = self.compiler1.compile_default_action()
        self.assertIsInstance(action_fluents, list)
        for fluent in action_fluents:
            self.assertIsInstance(fluent, tuple)
            self.assertEqual(len(fluent), 2)
            self.assertIsInstance(fluent[0], str)
            self.assertIsInstance(fluent[1], TensorFluent)

        af = dict(action_fluents)

        expected_action_fluents = {
            'outflow/1': { 'shape': [8,] , 'dtype': tf.float32 }
        }
        self.assertEqual(len(af), len(expected_action_fluents))
        for name, fluent in af.items():
            self.assertIn(name, expected_action_fluents)
            shape = expected_action_fluents[name]['shape']
            dtype = expected_action_fluents[name]['dtype']
            self.assertEqual(fluent.name, 'default_action/{}:0'.format(name.replace('/', '-')))
            self.assertIsInstance(fluent, TensorFluent)
            self.assertEqual(fluent.dtype, dtype)
            self.assertEqual(fluent.shape.as_list(), shape)

        expected_initializers = {
            'outflow/1': [0., 0., 0., 0., 0., 0., 0., 0.]
        }
        with tf.Session(graph=self.compiler1.graph) as sess:
            for name, fluent in af.items():
                value = sess.run(fluent.tensor)
                list1 = list(value.flatten())
                list2 = list(np.array(expected_initializers[name]).flatten())
                for v1, v2 in zip(list1, list2):
                    self.assertAlmostEqual(v1, v2)

    def test_state_scope(self):
        compilers = [self.compiler1, self.compiler2, self.compiler3, self.compiler4, self.compiler5, self.compiler6, self.compiler7]
        for compiler in compilers:
            fluents = compiler.compile_initial_state()
            scope = dict(fluents)
            self.assertEqual(len(fluents), len(scope))
            for i, name in enumerate(compiler.rddl.domain.state_fluent_ordering):
                self.assertIs(scope[name], fluents[i][1])

    def test_action_scope(self):
        compilers = [self.compiler1, self.compiler2, self.compiler3, self.compiler4, self.compiler5, self.compiler6, self.compiler7]
        for compiler in compilers:
            fluents = compiler.compile_default_action()
            scope = dict(fluents)
            self.assertEqual(len(fluents), len(scope))
            for i, name in enumerate(compiler.rddl.domain.action_fluent_ordering):
                self.assertIs(scope[name], fluents[i][1])

    def test_compile_expressions(self):
        expected = {
            # rddl1: RESERVOIR ====================================================
            'rainfall/1':   { 'shape': [8,], 'dtype': tf.float32, 'scope': ['?r'] },
            'evaporated/1': { 'shape': [8,], 'dtype': tf.float32, 'scope': ['?r'] },
            'overflow/1':   { 'shape': [8,], 'dtype': tf.float32, 'scope': ['?r'] },
            'inflow/1':     { 'shape': [8,], 'dtype': tf.float32, 'scope': ['?r'] },
            "rlevel'/1":    { 'shape': [8,], 'dtype': tf.float32, 'scope': ['?r'] },

            # rddl2: MARS ROVER ===================================================
            "xPos'/0":   { 'shape': [], 'dtype': tf.float32, 'scope': [] },
            "yPos'/0":   { 'shape': [], 'dtype': tf.float32, 'scope': [] },
            "time'/0":   { 'shape': [], 'dtype': tf.float32, 'scope': [] },
            "picTaken'/1": { 'shape': [3,], 'dtype': tf.bool, 'scope': ['?p'] }
        }

        compilers = [self.compiler1, self.compiler2]
        rddls = [self.rddl1, self.rddl2]
        for compiler, rddl in zip(compilers, rddls):
            nf = compiler.non_fluents_scope()
            sf = dict(compiler.compile_initial_state())
            af = dict(compiler.compile_default_action())
            scope = {}
            scope.update(nf)
            scope.update(sf)
            scope.update(af)

            _, cpfs = rddl.domain.cpfs
            for cpf in cpfs:
                name = cpf.name
                expr = cpf.expr
                t = compiler._compile_expression(expr, scope)
                scope[name] = t
                self.assertIsInstance(t, TensorFluent)
                self.assertEqual(t.shape.as_list(), expected[name]['shape'])
                self.assertEqual(t.dtype, expected[name]['dtype'])
                self.assertEqual(t.scope.as_list(), expected[name]['scope'])

            reward_expr = rddl.domain.reward
            t = compiler._compile_expression(reward_expr, scope)
            self.assertIsInstance(t, TensorFluent)
            self.assertEqual(t.shape.as_list(), [])
            self.assertEqual(t.dtype, tf.float32)
            self.assertEqual(t.scope.as_list(), [])

    def test_compile_cpfs(self):
        compilers = [self.compiler1, self.compiler2]
        expected = [
            (['evaporated/1', 'overflow/1', 'rainfall/1', 'inflow/1'], ["rlevel'/1"]),
            ([], ["picTaken'/1", "time'/0", "xPos'/0", "yPos'/0"]),
        ]
        for compiler, (expected_interm, expected_state) in zip(compilers, expected):
            nf = compiler.non_fluents_scope()
            sf = dict(compiler.compile_initial_state())
            af = dict(compiler.compile_default_action())
            scope = { **nf, **sf, **af }

            interm_fluents, next_state_fluents = compiler.compile_cpfs(scope)

            self.assertIsInstance(interm_fluents, list)
            self.assertEqual(len(interm_fluents), len(expected_interm))
            for fluent, expected_fluent in zip(interm_fluents, expected_interm):
                self.assertEqual(fluent[0], expected_fluent)

            self.assertIsInstance(next_state_fluents, list)
            self.assertEqual(len(next_state_fluents), len(expected_state))
            for fluent, expected_fluent in zip(next_state_fluents, expected_state):
                self.assertEqual(fluent[0], expected_fluent)

    def test_compile_state_cpfs(self):
        compilers = [self.compiler1, self.compiler2, self.compiler3, self.compiler4, self.compiler5, self.compiler6, self.compiler7]
        for compiler in compilers:
            nf = compiler.non_fluents_scope()
            sf = dict(compiler.compile_initial_state())
            af = dict(compiler.compile_default_action())
            scope = { **nf, **sf, **af }

            interm_fluents = compiler.compile_intermediate_cpfs(scope)
            scope.update(dict(interm_fluents))
            next_state_fluents = compiler.compile_state_cpfs(scope)

            self.assertIsInstance(next_state_fluents, list)
            for cpf in next_state_fluents:
                self.assertIsInstance(cpf, tuple)
            self.assertEqual(len(next_state_fluents), len(sf))

            next_state_fluents = dict(next_state_fluents)
            for fluent in sf:
                next_fluent = utils.rename_state_fluent(fluent)
                self.assertIn(next_fluent, next_state_fluents)
                self.assertIsInstance(next_state_fluents[next_fluent], TensorFluent)

    def test_compile_intermediate_cpfs(self):
        compilers = [self.compiler1, self.compiler2, self.compiler3, self.compiler4, self.compiler5, self.compiler6, self.compiler7]
        for compiler in compilers:
            fluents = compiler.rddl.domain.interm_fluent_ordering

            nf = compiler.non_fluents_scope()
            sf = dict(compiler.compile_initial_state())
            af = dict(compiler.compile_default_action())
            scope = { **nf, **sf, **af }
            interm_fluents = compiler.compile_intermediate_cpfs(scope)
            self.assertIsInstance(interm_fluents, list)
            self.assertEqual(len(interm_fluents), len(fluents))
            for actual, expected in zip(interm_fluents, fluents):
                self.assertIsInstance(actual, tuple)
                self.assertEqual(len(actual), 2)
                self.assertIsInstance(actual[0], str)
                self.assertIsInstance(actual[1], TensorFluent)
                self.assertEqual(actual[0], expected)

    def test_compile_reward(self):
        # TODO: self.compiler4
        compilers = [self.compiler1, self.compiler2, self.compiler3, self.compiler5, self.compiler6, self.compiler7]
        batch_size = 32
        for compiler in compilers:
            compiler.batch_mode_on()
            state = compiler.compile_initial_state(batch_size)
            action = compiler.compile_default_action(batch_size)
            scope = compiler.transition_scope(state, action)
            interm_fluents, next_state_fluents = compiler.compile_cpfs(scope)
            scope.update(next_state_fluents)
            reward = compiler.compile_reward(scope)
            self.assertIsInstance(reward, TensorFluent)
            self.assertEqual(reward.shape.as_list(), [batch_size])

    def test_compile_probabilistic_normal_random_variable(self):
        mean = Expression(('number', 0.0))
        var = Expression(('number', 1.0))
        normal = Expression(('randomvar', ('Normal', (mean, var))))

        expressions = [normal]
        self._test_random_variable_expressions(expressions)

    def test_compile_probabilistic_gamma_random_variable(self):
        shape = Expression(('number', 5.0))
        scale = Expression(('number', 1.0))
        gamma = Expression(('randomvar', ('Gamma', (shape, scale))))

        expressions = [gamma]
        self._test_random_variable_expressions(expressions)

    def __get_batch_compiler_with_state_action_scope(self):
        compilers = [self.compiler2]
        batch_sizes = [8]

        for compiler, batch_size in zip(compilers, batch_sizes):
            compiler.batch_mode_on()

            nf = compiler.non_fluents_scope()
            sf = dict(compiler.compile_initial_state())
            af = dict(compiler.compile_default_action())
            scope = { **nf, **sf, **af }

            yield (compiler, batch_size, scope)

    def _test_random_variable_expressions(self, expressions):
        for compiler, batch_size, scope in self.__get_batch_compiler_with_state_action_scope():
            for expr in expressions:
                self._test_random_variable_expression(expr, compiler, scope, batch_size)

    def _test_random_variable_expression(self, expr, compiler, scope, batch_size):
        sample = compiler._compile_random_variable_expression(expr, scope, batch_size)
        self._test_sample_fluents(sample, batch_size)
        self._test_sample_fluent(sample)

    def _test_sample_fluents(self, sample, batch_size=None):
        self.assertIsInstance(sample, TensorFluent)
        if batch_size is not None:
            self.assertEqual(sample.shape[0], batch_size)

    def _test_sample_fluent(self, sample):
        self.assertTrue(sample.tensor.name.startswith('sample'), sample.tensor)

    def _test_conditional_sample(self, sample):
        inputs = sample.tensor.op.inputs
        self.assertEqual(len(inputs), 3)
        self.assertTrue(inputs[0].name.startswith('LogicalNot'), inputs[0])
        self.assertTrue(inputs[1].name.startswith('StopGradient'), inputs[1])
        self.assertTrue(inputs[2].name.startswith('sample'), inputs[2])



@classmethod
    def _sample_log_prob(cls, dist, sample):
        tensor = dist.log_prob(tf.stop_gradient(sample.tensor))
        scope = []
        batch = sample.batch
        return TensorFluent(tensor, scope, batch)

    @classmethod
    def _deterministic_log_prob(cls, fluent):
        tensor = tf.constant(0.0, shape=fluent.shape._shape)
        scope = []
        batch = fluent.batch
        return TensorFluent(tensor, scope, batch)

    @classmethod
    def _sum_log_probs(cls, *args):
        total_log_prob = None
        for arg in args:
            if arg is not None:
                if total_log_prob is None:
                    total_log_prob = arg
                else:
                    total_log_prob += arg
        return total_log_prob

    @classmethod
    def _condition_log_prob(cls, condition, true_case, false_case, condition_log_prob, true_case_log_prob, false_case_log_prob):
        if condition_log_prob is None and true_case_log_prob is None and false_case_log_prob is None:
            return None

        true_total_log_prob = cls._sum_log_probs(condition_log_prob, true_case_log_prob)
        if true_total_log_prob is None:
            true_total_log_prob = cls._deterministic_log_prob(true_case)

        false_total_log_prob = cls._sum_log_probs(condition_log_prob, false_case_log_prob)
        if false_total_log_prob is None:
            false_total_log_prob = cls._deterministic_log_prob(false_case)

        true = TensorFluent.constant(True, tf.bool)
        false = TensorFluent.constant(False, tf.bool)
        log_prob_fluent = (condition == true) * true_total_log_prob + (condition == false) * false_total_log_prob
        tensor = log_prob_fluent.tensor
        scope = []
        batch = condition.batch
        return TensorFluent(tensor, scope, batch)

    @classmethod
    def _aggregation_log_prob(cls, log_prob, fluent, vars_list):
        if log_prob is None:
            return None
        axis = TensorFluent._varslist2axis(fluent, vars_list)
        tensor = tf.reduce_sum(log_prob.tensor, axis=axis)
        scope = []
        batch = fluent.batch
        return TensorFluent(tensor, scope, batch)



# This file is part of rddl2tf.

# rddl2tf is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# rddl2tf is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with rddl2tf. If not, see <http://www.gnu.org/licenses/>.


from pyrddl.rddl import RDDL
from pyrddl.pvariable import PVariable
from pyrddl.expr import Expression

from rddl2tf.fluent import TensorFluent
from rddl2tf import utils

import numpy as np
import tensorflow as tf

from typing import Dict, List, Optional, Sequence, Tuple, Union

CPFPair = Tuple[str, TensorFluent]
CPFTriple = Tuple[str, TensorFluent, TensorFluent]
FluentList = List[Tuple[str, TensorFluent]]
Bounds = Tuple[Optional[TensorFluent], Optional[TensorFluent]]
Value = Union[bool, int, float]
ArgsList = Optional[List[str]]
InitializerPair = Tuple[Tuple[str, ArgsList], Value]
InitializerList = List[InitializerPair]


class Compiler(object):
    '''RDDL2TensorFlow compiler.

    This is the core component of rddl2tf package.

    Its API provides methods to compile RDDL fluents and expressions
    to TensorFlow tensors wrapped as :obj:`rddl2tf.fluent.TensorFluent` objects.
    It supports constants, random variables, functions and operators
    used in most RDDL expressions. Also, it can handle next state
    and intermediate fluent CPFs, and rewards and action constraints.

    Args:
        rddl (:obj:`pyrddl.rddl.RDDL`): The RDDL model.
        batch_mode (bool): The batch mode flag.

    Attributes:
        rddl (:obj:`pyrddl.rddl.RDDL`): The RDDL model.
        batch_mode (bool): The batch mode flag.
        graph (:obj:`tensorflow.python.framework.ops.Graph`): The computation graph.
    '''

    def __init__(self, rddl: RDDL, batch_mode: bool = False) -> None:
        self.rddl = rddl
        self.batch_mode = batch_mode
        self.graph = tf.Graph()

    def batch_mode_on(self):
        '''Sets on the batch mode flag.'''
        self.batch_mode = True

    def batch_mode_off(self):
        '''Sets off the batch mode flag.'''
        self.batch_mode = False

    def compile_non_fluents(self):
        '''Returns a tuple of tensors representing the non fluents.

        Returns:
            Sequence[tf.Tensor]: A tuple of tensors.
        '''
        with self.graph.as_default():
            with tf.name_scope('non_fluents'):
                self._initialize_non_fluents()
                return self.non_fluents

    def compile_initial_state(self, batch_size: Optional[int] = None) -> Sequence[tf.Tensor]:
        '''Returns a tuple of tensors representing the initial state fluents.

        Args:
            batch_size (Optional[int]): The batch size.

        Returns:
            Sequence[tf.Tensor]: A tuple of tensors.
        '''
        with self.graph.as_default():
            with tf.name_scope('initial_state'):
                self._initialize_initial_state_fluents()
                if batch_size is None:
                    return self.initial_state_fluents
                return self._compile_batch_fluents(self.initial_state_fluents, batch_size)

    def compile_default_action(self, batch_size: Optional[int] = None) -> Sequence[tf.Tensor]:
        '''Returns a tuple of tensors representing the default action fluents.

        Args:
            batch_size (int): The batch size.

        Returns:
            Sequence[tf.Tensor]: A tuple of tensors.
        '''
        with self.graph.as_default():
            with tf.name_scope('default_action'):
                self._initialize_default_action_fluents()
                if batch_size is None:
                    return self.default_action_fluents
                return self._compile_batch_fluents(self.default_action_fluents, batch_size)

    def cpfs(self,
             state: Sequence[tf.Tensor],
             action: Sequence[tf.Tensor]) -> Tuple[List[TensorFluent], List[TensorFluent]]:
        '''Compiles the intermediate and next state fluent CPFs given
        the current `state` and `action`.

        Args:
            state (Sequence[tf.Tensor]): A tuple of state tensors.
            action (Sequence[tf.Tensor]): A tuple of action tensors.

        Returns:
            Tuple[List[TensorFluent], List[TensorFluent]]: A pair of lists of TensorFluent
            representing the intermediate and state CPFs.
        '''
        scope = self.transition_scope(state, action)
        batch_size = int(state[0].shape[0])
        interm_fluents, next_state_fluents = self.compile_cpfs(scope, batch_size)
        interms = [fluent for _, fluent in interm_fluents]
        next_state = [fluent for _, fluent in next_state_fluents]
        return interms, next_state

    def reward(self,
               state: Sequence[tf.Tensor],
               action: Sequence[tf.Tensor],
               next_state: Sequence[tf.Tensor]) -> tf.Tensor:
        '''Compiles the reward function given the current `state`, `action` and
        `next_state`.

        Args:
            state (Sequence[tf.Tensor]): A tuple of current state tensors.
            action (Sequence[tf.Tensor]): A tuple of action tensors.
            next_state (Sequence[tf.Tensor]): A tuple of next state tensors.

        Returns:
            (:obj:`tf.Tensor`): A tensor representing the reward function.
        '''
        scope = self.reward_scope(state, action, next_state)
        r = self.compile_reward(scope).tensor
        with self.graph.as_default():
            with tf.name_scope('reward'):
                return tf.expand_dims(r, -1)

    def compile_cpfs(self,
                     scope: Dict[str, TensorFluent],
                     batch_size: Optional[int] = None) -> Tuple[List[CPFPair], List[CPFPair]]:
        '''Compiles the intermediate and next state fluent CPFs given the current `state` and `action` scope.

        Args:
            scope (Dict[str, :obj:`rddl2tf.fluent.TensorFluent`]): The fluent scope for CPF evaluation.
            batch_size (Optional[int]): The batch size.

        Returns:
            Tuple[List[CPFPair], List[CPFPair]]: A pair of lists of TensorFluent
            representing the intermediate and state CPFs.
        '''
        interm_fluents = self.compile_intermediate_cpfs(scope, batch_size)
        scope.update(dict(interm_fluents))
        next_state_fluents = self.compile_state_cpfs(scope, batch_size)
        return interm_fluents, next_state_fluents

    def compile_intermediate_cpfs(self,
                                  scope: Dict[str, TensorFluent],
                                  batch_size: Optional[int] = None) -> List[CPFPair]:
        '''Compiles the intermediate fluent CPFs given the current `state` and `action` scope.

        Args:
            scope (Dict[str, :obj:`rddl2tf.fluent.TensorFluent`]): The fluent scope for CPF evaluation.
            batch_size (Optional[int]): The batch size.

        Returns:
            A list of intermediate fluent CPFs compiled to :obj:`rddl2tf.fluent.TensorFluent`.
        '''
        interm_fluents = []
        with self.graph.as_default():
            with tf.name_scope('intermediate_cpfs'):
                for cpf in self.rddl.domain.intermediate_cpfs:
                    name_scope = utils.identifier(cpf.name)
                    with tf.name_scope(name_scope):
                        t = self._compile_expression(cpf.expr, scope, batch_size)
                    interm_fluents.append((cpf.name, t))
                    scope[cpf.name] = t
                return interm_fluents

    def compile_state_cpfs(self,
                           scope: Dict[str, TensorFluent],
                           batch_size: Optional[int] = None) -> List[CPFPair]:
        '''Compiles the next state fluent CPFs given the current `state` and `action` scope.

        Args:
            scope (Dict[str, :obj:`rddl2tf.fluent.TensorFluent`]): The fluent scope for CPF evaluation.
            batch_size (Optional[int]): The batch size.

        Returns:
            A list of state fluent CPFs compiled to :obj:`rddl2tf.fluent.TensorFluent`.
        '''
        next_state_fluents = []
        with self.graph.as_default():
            with tf.name_scope('state_cpfs'):
                for cpf in self.rddl.domain.state_cpfs:
                    name_scope = utils.identifier(cpf.name)
                    with tf.name_scope(name_scope):
                        t = self._compile_expression(cpf.expr, scope, batch_size)
                    next_state_fluents.append((cpf.name, t))
                key = lambda f: self.rddl.domain.next_state_fluent_ordering.index(f[0])
                next_state_fluents = sorted(next_state_fluents, key=key)
                return next_state_fluents

    def compile_reward(self, scope: Dict[str, TensorFluent]) -> TensorFluent:
        '''Compiles the reward function given the fluent `scope`.

        Args:
            scope (Dict[str, :obj:`rddl2tf.fluent.TensorFluent`]): The fluent scope for reward evaluation.

        Returns:
            A :obj:`rddl2tf.fluent.TensorFluent` representing the reward function.
        '''
        reward_expr = self.rddl.domain.reward
        with self.graph.as_default():
            with tf.name_scope('reward'):
                return self._compile_expression(reward_expr, scope)

    def compile_state_action_constraints(self,
            state: Sequence[tf.Tensor],
            action: Sequence[tf.Tensor]) -> List[TensorFluent]:
        '''Compiles the state-action constraints given current `state` and `action` fluents.

        Args:
            state (Sequence[tf.Tensor]): The current state fluents.
            action (Sequence[tf.Tensor]): The action fluents.

        Returns:
            A list of :obj:`rddl2tf.fluent.TensorFluent`.
        '''
        scope = self.transition_scope(state, action)
        constraints = []
        with self.graph.as_default():
            with tf.name_scope('state_action_constraints'):
                for p in self.rddl.domain.constraints:
                    fluent = self._compile_expression(p, scope)
                    constraints.append(fluent)
                return constraints

    def compile_action_preconditions(self,
            state: Sequence[tf.Tensor],
            action: Sequence[tf.Tensor]) -> List[TensorFluent]:
        '''Compiles the action preconditions given current `state` and `action` fluents.

        Args:
            state (Sequence[tf.Tensor]): The current state fluents.
            action (Sequence[tf.Tensor]): The action fluents.

        Returns:
            A list of :obj:`rddl2tf.fluent.TensorFluent`.
        '''
        scope = self.action_precondition_scope(state, action)
        preconds = []
        with self.graph.as_default():
            with tf.name_scope('action_preconditions'):
                for p in self.rddl.domain.preconds:
                    fluent = self._compile_expression(p, scope)
                    preconds.append(fluent)
                return preconds

    def compile_state_invariants(self,
            state: Sequence[tf.Tensor]) -> List[TensorFluent]:
        '''Compiles the state invarints given current `state` fluents.

        Args:
            state (Sequence[tf.Tensor]): The current state fluents.

        Returns:
            A list of :obj:`rddl2tf.fluent.TensorFluent`.
        '''
        scope = self.state_invariant_scope(state)
        invariants = []
        with self.graph.as_default():
            with tf.name_scope('state_invariants'):
                for p in self.rddl.domain.invariants:
                    fluent = self._compile_expression(p, scope)
                    invariants.append(fluent)
                return invariants

    def compile_action_preconditions_checking(self,
            state: Sequence[tf.Tensor],
            action: Sequence[tf.Tensor]) -> tf.Tensor:
        '''Combines the action preconditions into an applicability checking op.

        Args:
            state (Sequence[tf.Tensor]): The current state fluents.
            action (Sequence[tf.Tensor]): The action fluents.

        Returns:
            A boolean tensor for checking if `action` is application in `state`.
        '''
        with self.graph.as_default():
            with tf.name_scope('action_preconditions_checking'):
                preconds = self.compile_action_preconditions(state, action)
                all_preconds = tf.stack([p.tensor for p in preconds], axis=1)
                checking = tf.reduce_all(all_preconds, axis=1)
                return checking

    def compile_action_bound_constraints(self,
            state: Sequence[tf.Tensor]) -> Dict[str, Bounds]:
        '''Compiles all actions bounds for the given `state`.

        Args:
            state (Sequence[tf.Tensor]): The current state fluents.

        Returns:
            A mapping from action names to a pair of
            :obj:`rddl2tf.fluent.TensorFluent` representing
            its lower and upper bounds.
        '''
        scope = self.action_precondition_scope(state)

        lower_bounds = self.rddl.domain.action_lower_bound_constraints
        upper_bounds = self.rddl.domain.action_upper_bound_constraints

        with self.graph.as_default():
            with tf.name_scope('action_bound_constraints'):

                bounds = {}
                for name in self.rddl.domain.action_fluent_ordering:

                    lower_expr = lower_bounds.get(name)
                    lower = None
                    if lower_expr is not None:
                        with tf.name_scope('lower_bound'):
                            lower = self._compile_expression(lower_expr, scope)

                    upper_expr = upper_bounds.get(name)
                    upper = None
                    if upper_expr is not None:
                        with tf.name_scope('upper_bound'):
                            upper = self._compile_expression(upper_expr, scope)

                    bounds[name] = (lower, upper)

                return bounds

    def non_fluents_scope(self) -> Dict[str, TensorFluent]:
        '''Returns a partial scope with non-fluents.

        Returns:
            A mapping from non-fluent names to :obj:`rddl2tf.fluent.TensorFluent`.
        '''
        if self.__dict__.get('non_fluents') is None:
            self._initialize_non_fluents()
        return dict(self.non_fluents)

    def state_scope(self, state_fluents: Sequence[tf.Tensor]) -> Dict[str, TensorFluent]:
        '''Returns a partial scope with current state-fluents.

        Args:
            state_fluents (Sequence[tf.Tensor]): The current state fluents.

        Returns:
            A mapping from state fluent names to :obj:`rddl2tf.fluent.TensorFluent`.
        '''
        return dict(zip(self.rddl.domain.state_fluent_ordering, state_fluents))

    def action_scope(self, action_fluents: Sequence[tf.Tensor]) -> Dict[str, TensorFluent]:
        '''Returns a partial scope with current action-fluents.

        Args:
            action_fluents (Sequence[tf.Tensor]): The action fluents.

        Returns:
            A mapping from action fluent names to :obj:`rddl2tf.fluent.TensorFluent`.
        '''
        return dict(zip(self.rddl.domain.action_fluent_ordering, action_fluents))

    def next_state_scope(self, next_state_fluents: Sequence[tf.Tensor]) -> Dict[str, TensorFluent]:
        '''Returns a partial scope with current next state-fluents.

        Args:
            next_state_fluents (Sequence[tf.Tensor]): The next state fluents.

        Returns:
            A mapping from next state fluent names to :obj:`rddl2tf.fluent.TensorFluent`.
        '''
        return dict(zip(self.rddl.domain.next_state_fluent_ordering, next_state_fluents))

    def transition_scope(self,
        state: Sequence[tf.Tensor],
        action: Sequence[tf.Tensor]) -> Dict[str, TensorFluent]:
        '''Returns the complete transition fluent scope
        for the current `state` and `action` fluents.

        Args:
            state (Sequence[tf.Tensor]): The current state fluents.
            action (Sequence[tf.Tensor]): The action fluents.

        Returns:
            A mapping from fluent names to :obj:`rddl2tf.fluent.TensorFluent`.
        '''
        scope = {}
        scope.update(self.non_fluents_scope())
        scope.update(self.state_scope(state))
        scope.update(self.action_scope(action))
        return scope

    def reward_scope(self,
                     state: Sequence[tf.Tensor],
                     action: Sequence[tf.Tensor],
                     next_state: Sequence[tf.Tensor]) -> Dict[str, TensorFluent]:
        '''Returns the complete reward fluent scope for the
        current `state`, `action` fluents, and `next_state` fluents.

        Args:
            state (Sequence[tf.Tensor]): The current state fluents.
            action (Sequence[tf.Tensor]): The action fluents.
            next_state (Sequence[tf.Tensor]): The next state fluents.

        Returns:
            A mapping from fluent names to :obj:`rddl2tf.fluent.TensorFluent`.
        '''
        scope = {}
        scope.update(self.non_fluents_scope())
        scope.update(self.state_scope(state))
        scope.update(self.action_scope(action))
        scope.update(self.next_state_scope(next_state))
        return scope

    def state_invariant_scope(self, state: Sequence[tf.Tensor]):
        '''Returns the state invariant fluent scope for the current `state`.

        Args:
            state (Sequence[tf.Tensor]): The current state fluents.

        Returns:
            A mapping from fluent names to :obj:`rddl2tf.fluent.TensorFluent`.
        '''
        scope = {}
        scope.update(self.non_fluents_scope())
        scope.update(self.state_scope(state))
        return scope

    def action_precondition_scope(self,
            state: Sequence[tf.Tensor],
            action: Optional[Sequence[tf.Tensor]] = None) -> Dict[str, TensorFluent]:
        '''Returns the action precondition fluent scope
        for the current `state` and `action` fluents.

        Args:
            state (Sequence[tf.Tensor]): The current state fluents.
            action (Sequence[tf.Tensor]): The action fluents.

        Returns:
            A mapping from fluent names to :obj:`rddl2tf.fluent.TensorFluent`.
        '''
        scope = {}
        scope.update(self.non_fluents_scope())
        scope.update(self.state_scope(state))
        if action is not None:
            scope.update(self.action_scope(action))
        return scope

    def _initialize_pvariables(self,
            pvariables: Dict[str, PVariable],
            ordering: List[str],
            initializer: Optional[InitializerList] = None) -> List[Tuple[str, TensorFluent]]:
        '''Instantiates `pvariables` given an initialization list and
        returns a list of TensorFluents in the given `ordering`.

        Returns:
            List[Tuple[str, TensorFluent]]: A list of pairs of fluent name and fluent tensor.
        '''
        if initializer is not None:
            init = dict()
            for ((name, args), value) in initializer:
                arity = len(args) if args is not None else 0
                name = '{}/{}'.format(name, arity)
                init[name] = init.get(name, [])
                init[name].append((args, value))

        fluents = []

        for name in ordering:
            pvar = pvariables[name]
            shape = self.rddl._param_types_to_shape(pvar.param_types)
            dtype = utils.range_type_to_dtype(pvar.range)
            fluent = np.full(shape, pvar.default)

            if initializer is not None:
                for args, val in init.get(name, []):
                    if args is not None:
                        idx = []
                        for ptype, arg in zip(pvar.param_types, args):
                            idx.append(self.rddl.object_table[ptype]['idx'][arg])
                        idx = tuple(idx)
                        fluent[idx] = val
                    else:
                        fluent = val

            with self.graph.as_default():
                t = tf.constant(fluent, dtype=dtype, name=utils.identifier(name))
                scope = [None] * len(t.shape)
                fluent = TensorFluent(t, scope, batch=False)
                fluent_pair = (name, fluent)
                fluents.append(fluent_pair)

        return fluents

    def _initialize_non_fluents(self):
        '''Returns the non-fluents instantiated.'''
        non_fluents = self.rddl.domain.non_fluents
        initializer = self.rddl.non_fluents.init_non_fluent
        self.non_fluents = self._initialize_pvariables(
            non_fluents,
            self.rddl.domain.non_fluent_ordering,
            initializer)
        return self.non_fluents

    def _initialize_initial_state_fluents(self):
        '''Returns the initial state-fluents instantiated.'''
        state_fluents = self.rddl.domain.state_fluents
        initializer = self.rddl.instance.init_state
        self.initial_state_fluents = self._initialize_pvariables(
            state_fluents,
            self.rddl.domain.state_fluent_ordering,
            initializer)
        return self.initial_state_fluents

    def _initialize_default_action_fluents(self):
        '''Returns the default action-fluents instantiated.'''
        action_fluents = self.rddl.domain.action_fluents
        self.default_action_fluents = self._initialize_pvariables(
            action_fluents,
            self.rddl.domain.action_fluent_ordering)
        return self.default_action_fluents

    def _compile_batch_fluents(self,
            fluents: List[Tuple[str, TensorFluent]],
            batch_size: int) -> Sequence[tf.Tensor]:
        '''Compiles `fluents` into tensors with given `batch_size`.

        Returns:
            Sequence[tf.Tensor]: A tuple of tensors with first dimension
            corresponding to the batch size.
        '''
        batch_fluents = []
        with self.graph.as_default():
            for name, fluent in fluents:
                name_scope = utils.identifier(name)
                with tf.name_scope(name_scope):
                    t = tf.stack([fluent.tensor] * batch_size)
                batch_fluents.append(t)
        return tuple(batch_fluents)

    def _compile_expression(self,
                            expr: Expression,
                            scope: Dict[str, TensorFluent],
                            batch_size: Optional[int] = None) -> TensorFluent:
        '''Compile the expression `expr` into a TensorFluent
        in the given `scope` with optional batch size.

        Args:
            expr (:obj:`rddl2tf.expr.Expression`): A RDDL expression.
            scope (Dict[str, :obj:`rddl2tf.fluent.TensorFluent`]): A fluent scope.
            batch_size (Optional[size]): The batch size.

        Returns:
            :obj:`rddl2tf.fluent.TensorFluent`: The compiled TensorFluent.
        '''
        etype = expr.etype

        with self.graph.as_default():

            if etype[0] == 'constant':
                return self._compile_constant_expression(expr, scope, batch_size)
            elif etype[0] == 'pvar':
                return self._compile_pvariable_expression(expr, scope, batch_size)
            elif etype[0] == 'randomvar':
                return self._compile_random_variable_expression(expr, scope, batch_size)
            elif etype[0] == 'arithmetic':
                return self._compile_arithmetic_expression(expr, scope, batch_size)
            elif etype[0] == 'boolean':
                return self._compile_boolean_expression(expr, scope, batch_size)
            elif etype[0] == 'relational':
                return self._compile_relational_expression(expr, scope, batch_size)
            elif etype[0] == 'func':
                return self._compile_function_expression(expr, scope, batch_size)
            elif etype[0] == 'control':
                return self._compile_control_flow_expression(expr, scope, batch_size)
            elif etype[0] == 'aggregation':
                return self._compile_aggregation_expression(expr, scope, batch_size)
            else:
                raise ValueError('Expression type unknown: {}'.format(etype))

    def _compile_constant_expression(self,
                                     expr: Expression,
                                     scope: Dict[str, TensorFluent],
                                     batch_size: Optional[int] = None) -> TensorFluent:
        '''Compile a constant expression `expr` into a TensorFluent
        in the given `scope` with optional batch size.

        Args:
            expr (:obj:`rddl2tf.expr.Expression`): A RDDL constant expression.
            scope (Dict[str, :obj:`rddl2tf.fluent.TensorFluent`]): A fluent scope.
            batch_size (Optional[size]): The batch size.

        Returns:
            :obj:`rddl2tf.fluent.TensorFluent`: The compiled expression as a TensorFluent.
        '''
        etype = expr.etype
        args = expr.args
        dtype = utils.python_type_to_dtype(etype[1])
        fluent = TensorFluent.constant(args, dtype=dtype)
        return fluent

    def _compile_pvariable_expression(self,
                                      expr: Expression,
                                      scope: Dict[str, TensorFluent],
                                      batch_size: Optional[int] = None) -> TensorFluent:
        '''Compile a pvariable expression `expr` into a TensorFluent
        in the given `scope` with optional batch size.

        Args:
            expr (:obj:`rddl2tf.expr.Expression`): A RDDL pvariable expression.
            scope (Dict[str, :obj:`rddl2tf.fluent.TensorFluent`]): A fluent scope.
            batch_size (Optional[size]): The batch size.

        Returns:
            :obj:`rddl2tf.fluent.TensorFluent`: The compiled expression as a TensorFluent.
        '''
        etype = expr.etype
        args = expr.args
        name = expr._pvar_to_name(args)
        if name not in scope:
            raise ValueError('Variable {} not in scope.'.format(name))
        fluent = scope[name]
        scope = args[1] if args[1] is not None else []
        if isinstance(fluent, TensorFluent):
            fluent = TensorFluent(fluent.tensor, scope, batch=fluent.batch)
        elif isinstance(fluent, tf.Tensor):
            fluent = TensorFluent(fluent, scope, batch=self.batch_mode)
        else:
            raise ValueError('Variable in scope must be TensorFluent-like: {}'.format(fluent))
        return fluent

    def _compile_random_variable_expression(self,
                                            expr: Expression,
                                            scope: Dict[str, TensorFluent],
                                            batch_size: Optional[int] = None) -> TensorFluent:
        '''Compile a random variable expression `expr` into a TensorFluent
        in the given `scope` with optional batch size.

        If `reparam` tensor is given, then it conditionally stops gradient
        backpropagation at the batch level where `reparam` is False.

        Args:
            expr (:obj:`rddl2tf.expr.Expression`): A RDDL random variable expression.
            scope (Dict[str, :obj:`rddl2tf.fluent.TensorFluent`]): A fluent scope.
            batch_size (Optional[size]): The batch size.

        Returns:
            :obj:`rddl2tf.fluent.TensorFluent`: The compiled expression as a TensorFluent.
        '''
        etype = expr.etype
        args = expr.args

        if etype[1] == 'KronDelta':
            sample = self._compile_expression(args[0], scope)
        elif etype[1] == 'Bernoulli':
            mean = self._compile_expression(args[0], scope)
            dist, sample = TensorFluent.Bernoulli(mean, batch_size)
        elif etype[1] == 'Uniform':
            low = self._compile_expression(args[0], scope)
            high = self._compile_expression(args[1], scope)
            dist, sample = TensorFluent.Uniform(low, high, batch_size)
        elif etype[1] == 'Normal':
            mean = self._compile_expression(args[0], scope)
            variance = self._compile_expression(args[1], scope)
            dist, sample = TensorFluent.Normal(mean, variance, batch_size)
        elif etype[1] == 'Laplace':
            mean = self._compile_expression(args[0], scope)
            variance = self._compile_expression(args[1], scope)
            dist, sample = TensorFluent.Laplace(mean, variance, batch_size)
        elif etype[1] == 'Gamma':
            shape = self._compile_expression(args[0], scope)
            scale = self._compile_expression(args[1], scope)
            dist, sample = TensorFluent.Gamma(shape, scale, batch_size)
        elif etype[1] == 'Exponential':
            mean = self._compile_expression(args[0], scope)
            dist, sample = TensorFluent.Exponential(mean, batch_size)
        else:
            raise ValueError('Invalid random variable expression:\n{}.'.format(expr))

        return sample

    def _compile_arithmetic_expression(self,
                                       expr: Expression,
                                       scope: Dict[str, TensorFluent],
                                       batch_size: Optional[int] = None) -> TensorFluent:
        '''Compile an arithmetic expression `expr` into a TensorFluent
        in the given `scope` with optional batch size.

        Args:
            expr (:obj:`rddl2tf.expr.Expression`): A RDDL arithmetic expression.
            scope (Dict[str, :obj:`rddl2tf.fluent.TensorFluent`]): A fluent scope.
            batch_size (Optional[size]): The batch size.

        Returns:
            :obj:`rddl2tf.fluent.TensorFluent`: The compiled expression as a TensorFluent.
        '''
        etype = expr.etype
        args = expr.args

        if len(args) == 1:
            if etype[1] == '+':
                fluent = self._compile_expression(args[0], scope)
            elif etype[1] == '-':
                op1 = self._compile_expression(args[0], scope)
                fluent = -op1
            else:
                raise ValueError('Invalid unary arithmetic expression:\n{}'.format(expr))
        else:
            if etype[1] == '+':
                op1 = self._compile_expression(args[0], scope)
                op2 = self._compile_expression(args[1], scope)
                fluent = op1 + op2
            elif etype[1] == '-':
                op1 = self._compile_expression(args[0], scope)
                op2 = self._compile_expression(args[1], scope)
                fluent = op1 - op2
            elif etype[1] == '*':
                op1 = self._compile_expression(args[0], scope)
                op2 = self._compile_expression(args[1], scope)
                fluent = op1 * op2
            elif etype[1] == '/':
                op1 = self._compile_expression(args[0], scope)
                op2 = self._compile_expression(args[1], scope)
                fluent = op1 / op2
            else:
                raise ValueError('Invalid binary arithmetic expression:\n{}'.format(expr))

        return fluent

    def _compile_boolean_expression(self,
                                    expr: Expression,
                                    scope: Dict[str, TensorFluent],
                                    batch_size: Optional[int] = None) -> TensorFluent:
        '''Compile a boolean/logical expression `expr` into a TensorFluent
        in the given `scope` with optional batch size.

        Args:
            expr (:obj:`rddl2tf.expr.Expression`): A RDDL boolean expression.
            scope (Dict[str, :obj:`rddl2tf.fluent.TensorFluent`]): A fluent scope.
            batch_size (Optional[size]): The batch size.

        Returns:
            :obj:`rddl2tf.fluent.TensorFluent`: The compiled expression as a TensorFluent.
        '''
        etype = expr.etype
        args = expr.args

        if len(args) == 1:
            if etype[1] == '~':
                op = self._compile_expression(args[0], scope)
                fluent = ~op
            else:
                raise ValueError('Invalid unary boolean expression:\n{}'.format(expr))
        else:
            if etype[1] in ['^', '&']:
                op1 = self._compile_expression(args[0], scope)
                op2 = self._compile_expression(args[1], scope)
                fluent = op1 & op2
            elif etype[1] == '|':
                op1 = self._compile_expression(args[0], scope)
                op2 = self._compile_expression(args[1], scope)
                fluent = op1 | op2
            elif etype[1] == '=>':
                op1 = self._compile_expression(args[0], scope)
                op2 = self._compile_expression(args[1], scope)
                fluent = ~op1 | op2
            elif etype[1] == '<=>':
                op1 = self._compile_expression(args[0], scope)
                op2 = self._compile_expression(args[1], scope)
                fluent = (op1 & op2) | (~op1 & ~op2)
            else:
                raise ValueError('Invalid binary boolean expression:\n{}'.format(expr))

        return fluent

    def _compile_relational_expression(self,
                                       expr: Expression,
                                       scope: Dict[str, TensorFluent],
                                       batch_size: Optional[int] = None) -> TensorFluent:
        '''Compile a relational expression `expr` into a TensorFluent
        in the given `scope` with optional batch size.

        Args:
            expr (:obj:`rddl2tf.expr.Expression`): A RDDL relational expression.
            scope (Dict[str, :obj:`rddl2tf.fluent.TensorFluent`]): A fluent scope.
            batch_size (Optional[size]): The batch size.

        Returns:
            :obj:`rddl2tf.fluent.TensorFluent`: The compiled expression as a TensorFluent.
        '''
        etype = expr.etype
        args = expr.args

        if etype[1] == '<=':
            op1 = self._compile_expression(args[0], scope)
            op2 = self._compile_expression(args[1], scope)
            fluent = (op1 <= op2)
        elif etype[1] == '<':
            op1 = self._compile_expression(args[0], scope)
            op2 = self._compile_expression(args[1], scope)
            fluent = (op1 < op2)
        elif etype[1] == '>=':
            op1 = self._compile_expression(args[0], scope)
            op2 = self._compile_expression(args[1], scope)
            fluent = (op1 >= op2)
        elif etype[1] == '>':
            op1 = self._compile_expression(args[0], scope)
            op2 = self._compile_expression(args[1], scope)
            fluent = (op1 > op2)
        elif etype[1] == '==':
            op1 = self._compile_expression(args[0], scope)
            op2 = self._compile_expression(args[1], scope)
            fluent = (op1 == op2)
        elif etype[1] == '~=':
            op1 = self._compile_expression(args[0], scope)
            op2 = self._compile_expression(args[1], scope)
            fluent = (op1 != op2)
        else:
            raise ValueError('Invalid relational expression:\n{}'.format(expr))

        return fluent

    def _compile_function_expression(self,
                                     expr: Expression,
                                     scope: Dict[str, TensorFluent],
                                     batch_size: Optional[int] = None) -> TensorFluent:
        '''Compile a function expression `expr` into a TensorFluent
        in the given `scope` with optional batch size.

        Args:
            expr (:obj:`rddl2tf.expr.Expression`): A RDDL function expression.
            scope (Dict[str, :obj:`rddl2tf.fluent.TensorFluent`]): A fluent scope.
            batch_size (Optional[size]): The batch size.

        Returns:
            :obj:`rddl2tf.fluent.TensorFluent`: The compiled expression as a TensorFluent.
        '''
        etype = expr.etype
        args = expr.args

        if len(args) == 1:
            if etype[1] == 'abs':
                x = self._compile_expression(args[0], scope)
                fluent = TensorFluent.abs(x)
            elif etype[1] == 'exp':
                x = self._compile_expression(args[0], scope)
                fluent = TensorFluent.exp(x)
            elif etype[1] == 'log':
                x = self._compile_expression(args[0], scope)
                fluent = TensorFluent.log(x)
            elif etype[1] == 'sqrt':
                x = self._compile_expression(args[0], scope)
                fluent = TensorFluent.sqrt(x)
            elif etype[1] == 'cos':
                x = self._compile_expression(args[0], scope)
                fluent = TensorFluent.cos(x)
            elif etype[1] == 'sin':
                x = self._compile_expression(args[0], scope)
                fluent = TensorFluent.sin(x)
            elif etype[1] == 'tan':
                x = self._compile_expression(args[0], scope)
                fluent = TensorFluent.tan(x)
            elif etype[1] in ['acos', 'arccos']:
                x = self._compile_expression(args[0], scope)
                fluent = TensorFluent.acos(x)
            elif etype[1] in ['asin', 'arcsin']:
                x = self._compile_expression(args[0], scope)
                fluent = TensorFluent.asin(x)
            elif etype[1] in ['atan', 'arctan']:
                x = self._compile_expression(args[0], scope)
                fluent = TensorFluent.atan(x)
            elif etype[1] == 'round':
                x = self._compile_expression(args[0], scope)
                fluent = TensorFluent.round(x)
            elif etype[1] == 'ceil':
                x = self._compile_expression(args[0], scope)
                fluent = TensorFluent.ceil(x)
            elif etype[1] == 'floor':
                x = self._compile_expression(args[0], scope)
                fluent = TensorFluent.floor(x)
            else:
                raise ValueError('Invalid unary function expression:\n{}'.format(expr))
        else:
            if etype[1] == 'pow':
                x = self._compile_expression(args[0], scope)
                y = self._compile_expression(args[1], scope)
                fluent = TensorFluent.pow(x, y)
            elif etype[1] == 'max':
                x = self._compile_expression(args[0], scope)
                y = self._compile_expression(args[1], scope)
                fluent = TensorFluent.max(x, y)
            elif etype[1] == 'min':
                x = self._compile_expression(args[0], scope)
                y = self._compile_expression(args[1], scope)
                fluent = TensorFluent.min(x, y)
            else:
                raise ValueError('Invalid binary function expression:\n{}'.format(expr))

        return fluent

    def _compile_control_flow_expression(self,
                                         expr: Expression,
                                         scope: Dict[str, TensorFluent],
                                         batch_size: Optional[int] = None) -> TensorFluent:
        '''Compile a control flow expression `expr` into a TensorFluent
        in the given `scope` with optional batch size.

        Args:
            expr (:obj:`rddl2tf.expr.Expression`): A RDDL control flow expression.
            scope (Dict[str, :obj:`rddl2tf.fluent.TensorFluent`]): A fluent scope.
            batch_size (Optional[size]): The batch size.

        Returns:
            :obj:`rddl2tf.fluent.TensorFluent`: The compiled expression as a TensorFluent.
        '''
        etype = expr.etype
        args = expr.args
        if etype[1] == 'if':
            condition = self._compile_expression(args[0], scope)
            true_case = self._compile_expression(args[1], scope)
            false_case = self._compile_expression(args[2], scope)
            fluent = TensorFluent.if_then_else(condition, true_case, false_case)
        else:
            raise ValueError('Invalid control flow expression:\n{}'.format(expr))
        return fluent

    def _compile_aggregation_expression(self,
                                        expr: Expression,
                                        scope: Dict[str, TensorFluent],
                                        batch_size: Optional[int] = None) -> TensorFluent:
        '''Compile an aggregation expression `expr` into a TensorFluent
        in the given `scope` with optional batch size.

        Args:
            expr (:obj:`rddl2tf.expr.Expression`): A RDDL aggregation expression.
            scope (Dict[str, :obj:`rddl2tf.fluent.TensorFluent`]): A fluent scope.
            batch_size (Optional[size]): The batch size.

        Returns:
            :obj:`rddl2tf.fluent.TensorFluent`: The compiled expression as a TensorFluent.
        '''
        etype = expr.etype
        args = expr.args

        typed_var_list = args[:-1]
        vars_list = [var for _, (var, _) in typed_var_list]
        expr = args[-1]

        x = self._compile_expression(expr, scope)

        if etype[1] == 'sum':
            fluent = x.sum(vars_list=vars_list)
        elif etype[1] == 'prod':
            fluent = x.prod(vars_list=vars_list)
        elif etype[1] == 'avg':
            fluent = x.avg(vars_list=vars_list)
        elif etype[1] == 'maximum':
            fluent = x.maximum(vars_list=vars_list)
        elif etype[1] == 'minimum':
            fluent = x.minimum(vars_list=vars_list)
        elif etype[1] == 'exists':
            fluent = x.exists(vars_list=vars_list)
        elif etype[1] == 'forall':
            fluent = x.forall(vars_list=vars_list)
        else:
            raise ValueError('Invalid aggregation expression {}.'.format(expr))

        return fluent



# This file is part of rddl2tf.

# rddl2tf is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# rddl2tf is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with rddl2tf. If not, see <http://www.gnu.org/licenses/>.


from rddl2tf.fluent import TensorFluent

import numpy as np
import tensorflow as tf

import unittest


class TestTensorFluent(unittest.TestCase):

    def setUp(self):
        self.batch_size = 32
        tf.reset_default_graph()
        self.zero = TensorFluent.constant(0.0)
        self.one = TensorFluent.constant(1)

    def test_constant_fluent(self):
        self._test_op_name(self.zero, r'^Const')
        self._test_op_name(self.one, r'^Const')
        self._test_fluent(self.zero, tf.float32, [], [], False)
        self._test_fluent(self.one, tf.float32, [], [], False)

    def test_normal_fluent(self):
        mean = self.zero
        variance = self.one
        dist, fluent = TensorFluent.Normal(mean, variance, self.batch_size)
        self.assertIsInstance(dist, tf.distributions.Normal)
        self._test_op_name(fluent, r'^Normal[^/]*/sample')
        self._test_fluent(fluent, tf.float32, [self.batch_size], [], True)

    def test_if_then_else(self):
        # if (abs[x - gx] > 0.0) then
        #     (u * u) * 0.01
        # else
        #     0.0
        x = tf.random_normal([self.batch_size, 1])
        x = TensorFluent(x, scope=[], batch=True)
        gx = tf.random_normal([])
        gx = TensorFluent(gx, scope=[], batch=False)
        u = tf.random_normal([self.batch_size,])
        u = TensorFluent(u, scope=[], batch=True)
        c0 = TensorFluent.constant(0.01)
        cond = TensorFluent.abs(x - gx) > self.zero
        true_case = (u * u) * c0
        false_case = self.zero
        ite = TensorFluent.if_then_else(cond, true_case, false_case)
        self._test_fluent(ite, tf.float32, [self.batch_size, 1], [], True)

        # if ((u * u) * 0.01 > 1.0) then
        #     abs[x - gx]
        # else
        #     0.0
        cond = ((u * u) * c0) > self.one
        true_case = TensorFluent.abs(x - gx)
        false_case = self.zero
        ite = TensorFluent.if_then_else(cond, true_case, false_case)
        self._test_fluent(ite, tf.float32, [self.batch_size, 1], [], True)

        # if (gx < 1.0) then
        #     1.0 + u
        # else
        #     0.05 * (u * u)
        c1 = TensorFluent.constant(0.05)
        cond = (gx < self.one)
        true_case = self.one + u
        false_case = c1 * (u * u)
        ite = TensorFluent.if_then_else(cond, true_case, false_case)
        self._test_fluent(ite, tf.float32, [self.batch_size], [], True)

        # if (gx < 0.0) then
        #     1.0
        # else
        #     abs[x - gx]
        cond = (gx < self.zero)
        true_case = self.one
        false_case = TensorFluent.abs(x - gx)
        ite = TensorFluent.if_then_else(cond, true_case, false_case)
        self._test_fluent(ite, tf.float32, [self.batch_size, 1], [], True)

        # if (visited(?wpt)) then
        #     (1.0)
        # else
        #     (0.0)
        visited = tf.stack([[True, False, True]] * self.batch_size, axis=0)
        visited = TensorFluent(visited, scope=['?wpt'], batch=True)
        cond = visited
        true_case = self.one
        false_case = self.zero
        ite = TensorFluent.if_then_else(cond, true_case, false_case)
        self._test_fluent(ite, tf.float32, [self.batch_size, 3], ['?wpt'], True)

        # if (rlevel'(?r)<=LOWER_BOUND(?r)) then
        #     LOW_PENALTY(?r)*(LOWER_BOUND(?r)-rlevel'(?r))
        # else
        #     HIGH_PENALTY(?r)*(rlevel'(?r)-UPPER_BOUND(?r))];
        r_size = 8
        low_penalty = TensorFluent.constant(-5.0)
        high_penalty = TensorFluent.constant(-10.0)
        lower_bound = tf.random_normal([r_size], mean=50.0, stddev=10.0)
        lower_bound = TensorFluent(lower_bound, scope=['?r'], batch=False)
        upper_bound = tf.random_normal([r_size], mean=200.0, stddev=10.0)
        upper_bound = TensorFluent(upper_bound, scope=['?r'], batch=False)
        rlevel = tf.random_normal([self.batch_size, r_size], mean=100.0, stddev=10.0)
        rlevel = TensorFluent(rlevel, scope=['?r'], batch=True)
        cond = (rlevel <= lower_bound)
        true_case = low_penalty * (lower_bound - rlevel)
        false_case = high_penalty * (rlevel - upper_bound)
        ite = TensorFluent.if_then_else(cond, true_case, false_case)
        self._test_fluent(ite, tf.float32, [self.batch_size, r_size], ['?r'], True)

        # if (snapPicture) then
        #    (time + 0.25)
        # else
        #    (time + abs[xMove] + abs[yMove]);
        time = tf.fill([self.batch_size, 1], 12.0)
        time = TensorFluent(time, scope=[], batch=True)
        snapPicture = tf.fill([self.batch_size], True)
        snapPicture = TensorFluent(snapPicture, scope=[], batch=True)
        xMove = tf.fill([self.batch_size], 10.0)
        xMove = TensorFluent(xMove, scope=[], batch=True)
        yMove = tf.fill([self.batch_size], -3.0)
        yMove = TensorFluent(yMove, scope=[], batch=True)
        cond = snapPicture
        true_case = (time + TensorFluent.constant(0.25))
        false_case = (time + TensorFluent.abs(xMove) + TensorFluent.abs(yMove))
        ite = TensorFluent.if_then_else(cond, true_case, false_case)
        self._test_fluent(ite, tf.float32, [self.batch_size, 1], [], True)

        # if (alive(?x,?y)) then
        #     Bernoulli(1.0 - NOISE-PROB(?x,?y))
        # else
        #     Bernoulli(NOISE-PROB(?x,?y));
        x_size, y_size = 3, 3
        alive = tf.distributions.Bernoulli(probs=0.7, dtype=tf.bool).\
            sample([self.batch_size, x_size, y_size])
        alive = TensorFluent(alive, scope=['?x', '?y'], batch=True)
        noise_prob = tf.random_uniform([x_size, y_size], dtype=tf.float32)
        noise_prob = TensorFluent(noise_prob, scope=['?x', '?y'], batch=False)
        cond = alive
        true_case = TensorFluent.Bernoulli(self.one - noise_prob, batch_size=self.batch_size)[1]
        false_case = TensorFluent.Bernoulli(noise_prob, batch_size=self.batch_size)[1]
        ite = TensorFluent.if_then_else(cond, true_case, false_case)
        self._test_fluent(ite, tf.bool, [self.batch_size, 3, 3], ['?x', '?y'], True)

        # if (reboot(?x)) then
        #     KronDelta(true)
        # else
        #     ~running(?x);
        x_size = 128
        mean = tf.random_uniform([x_size], dtype=tf.float32)
        mean = TensorFluent(mean, scope=['?x'])
        running = TensorFluent.Bernoulli(mean, batch_size=self.batch_size)[1]
        reboot = TensorFluent.Bernoulli(self.one - mean, batch_size=self.batch_size)[1]
        cond = reboot
        true_case = TensorFluent.constant(True, dtype=tf.bool)
        false_case = ~running
        ite = TensorFluent.if_then_else(cond, true_case, false_case)
        self._test_fluent(ite, tf.bool, [self.batch_size, x_size], ['?x'], True)

        # if (running(?x)) then
        #     Bernoulli(.45 + .5 * running(?x))
        # else
        #     Bernoulli(REBOOT-PROB);
        c0 = TensorFluent.constant(0.45)
        c1 = TensorFluent.constant(0.5)
        reboot_prob = TensorFluent.constant(0.15)
        cond = running
        true_case = TensorFluent.Bernoulli(c0 + c1 * running)[1]
        false_case = TensorFluent.Bernoulli(reboot_prob)[1]
        ite = TensorFluent.if_then_else(cond, true_case, false_case)
        self._test_fluent(ite, tf.bool, [self.batch_size, x_size], ['?x'], True)

    def test_stop_gradient(self):
        mean = self.zero
        variance = self.one
        _, sample = TensorFluent.Normal(mean, variance, self.batch_size)
        fluent = TensorFluent.stop_gradient(sample)
        self._test_fluent(fluent, sample.dtype, sample.shape.as_list(), sample.scope.as_list(), sample.batch)
        self._test_op_name(fluent, r'^StopGradient')

        grad_before, = tf.gradients(ys=tf.reduce_sum(sample.tensor), xs=mean.tensor)
        grad_after, = tf.gradients(ys=tf.reduce_sum(fluent.tensor), xs=mean.tensor)

        self.assertIsInstance(grad_before, tf.Tensor)
        self.assertIsNone(grad_after)

        with tf.Session() as sess:
            f1, f2 = sess.run([sample.tensor, fluent.tensor])
            self.assertListEqual(list(f1), list(f2))

            g1 = sess.run(grad_before)
            self.assertEqual(g1, self.batch_size)

    def test_stop_batch_gradient(self):
        mean = TensorFluent(tf.zeros([self.batch_size]), [], True)
        variance = TensorFluent(tf.ones([self.batch_size]), [], True)
        _, sample = TensorFluent.Normal(mean, variance)

        stop_batch = tf.cast(tf.distributions.Bernoulli(probs=0.5).sample(self.batch_size), tf.bool)

        fluent = TensorFluent.stop_batch_gradient(sample, stop_batch)
        self._test_fluent(fluent, sample.dtype, sample.shape.as_list(), sample.scope.as_list(), sample.batch)
        self._test_op_name(fluent, r'^Select')

        grad_before, = tf.gradients(ys=tf.reduce_sum(sample.tensor), xs=mean.tensor)
        grad_after, = tf.gradients(ys=tf.reduce_sum(fluent.tensor), xs=mean.tensor)

        self.assertIsInstance(grad_before, tf.Tensor)
        self.assertIsInstance(grad_after, tf.Tensor)

        with tf.Session() as sess:
            f1, f2 = sess.run([sample.tensor, fluent.tensor])
            self.assertListEqual(list(f1), list(f2))

            g1, g2, stop = sess.run([grad_before, grad_after, stop_batch])
            self.assertTrue(all(g1 == np.ones([self.batch_size])))
            self.assertTrue(all(g2 == (~stop).astype(np.float32)))



    def _test_fluent(self, fluent, dtype, shape, scope, batch):
        self.assertIsInstance(fluent, TensorFluent)
        self.assertEqual(fluent.dtype, dtype)
        self.assertListEqual(fluent.shape.as_list(), shape)
        self.assertListEqual(fluent.scope.as_list(), scope)
        self.assertEqual(fluent.batch, batch)

    def _test_op_name(self, fluent, regex):
        self.assertRegex(fluent.tensor.op.name, regex, fluent.tensor)



# This file is part of rddl2tf.

# rddl2tf is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# rddl2tf is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with rddl2tf. If not, see <http://www.gnu.org/licenses/>.


from pyrddl.expr import Expression

from rddl2tf.fluent import TensorFluent
from rddl2tf.fluentshape import TensorFluentShape

import numpy as np
from typing import Dict, List, Tuple

Noise = List[Tuple[str, List[int]]]


def get_reparameterization(expr: Expression,
                           scope: Dict[str, TensorFluent]) -> Noise:
    noise = []
    _get_reparameterization(expr, scope, noise)
    return noise


def _get_reparameterization(expr: Expression,
                            scope: Dict[str, TensorFluent],
                            noise: List) -> TensorFluentShape:
    etype = expr.etype
    args = expr.args

    if etype[0] == 'constant':
        return TensorFluentShape([1], batch=False)
    elif etype[0] == 'pvar':
        name = expr._pvar_to_name(args)
        if name not in scope:
            raise ValueError('Variable {} not in scope.'.format(name))
        fluent = scope[name]
        return fluent.shape
    elif etype[0] == 'randomvar':
        if etype[1] == 'Normal':
            mean_shape = _get_reparameterization(args[0], scope, noise)
            var_shape = _get_reparameterization(args[1], scope, noise)
            shape = _broadcast(mean_shape, var_shape)
            noise.append(('Normal', shape.as_list()))
            return shape
    elif etype[0] in ['arithmetic', 'boolean', 'relational']:
        op1_shape = _get_reparameterization(args[0], scope, noise)
        shape = op1_shape
        if len(args) > 1:
            op2_shape = _get_reparameterization(args[1], scope, noise)
            shape = _broadcast(op1_shape, op2_shape)
        return shape
    elif etype[0] == 'func':
        op1_shape = _get_reparameterization(args[0], scope, noise)
        shape = op1_shape
        if len(args) > 1:
            if len(args) == 2:
                op2_shape = _get_reparameterization(args[1], scope, noise)
                shape = _broadcast(op1_shape, op2_shape)
            else:
                raise ValueError('Invalid function:\n{}'.format(expr))
        return shape
    elif etype[0] == 'control':
        if etype[1] == 'if':
            condition_shape = _get_reparameterization(args[0], scope, noise)
            true_case_shape = _get_reparameterization(args[1], scope, noise)
            false_case_shape = _get_reparameterization(args[2], scope, noise)
            shape = _broadcast(condition_shape, true_case_shape)
            shape = _broadcast(shape, false_case_shape)
            return shape
        else:
            raise ValueError('Invalid control flow expression:\n{}'.format(expr))
    elif etype[0] == 'aggregation':
        return _get_reparameterization(args[-1], scope, noise)

    raise ValueError('Expression type unknown: {}'.format(etype))


def _broadcast(shape1: TensorFluentShape, shape2: TensorFluentShape) -> TensorFluentShape:
    s1, s2 = TensorFluentShape.broadcast(shape1, shape2)
    s1 = s1 if s1 is not None else shape1.as_list()
    s2 = s2 if s2 is not None else shape2.as_list()
    x1, x2 = np.zeros(s1), np.zeros(s2)
    y = np.broadcast(x1, x2)
    return TensorFluentShape(y.shape, batch=(shape1.batch or shape2.batch))




# This file is part of rddl2tf.

# rddl2tf is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# rddl2tf is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with rddl2tf. If not, see <http://www.gnu.org/licenses/>.


from typing import Optional

import tensorflow as tf


def range_type_to_dtype(range_type: str) -> Optional[tf.DType]:
    '''Maps RDDL range types to TensorFlow dtypes.'''
    range2dtype = {
        'real': tf.float32,
        'int': tf.int32,
        'bool': tf.bool
    }
    return range2dtype[range_type]


def python_type_to_dtype(python_type: type) -> Optional[tf.DType]:
    '''Maps python types to TensorFlow dtypes.'''
    dtype = None
    if python_type == float:
        dtype = tf.float32
    elif python_type == int:
        dtype = tf.int32
    elif python_type == bool:
        dtype = tf.bool
    return dtype


def identifier(name):
    name = name.replace("'", '')
    name = name.replace('/', '-')
    return name



# This file is part of rddl2tf.

# rddl2tf is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# rddl2tf is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with rddl2tf. If not, see <http://www.gnu.org/licenses/>.


from pyrddl.expr import Expression

from rddl2tf.fluent import TensorFluent
from rddl2tf.reparam import get_reparameterization

import tensorflow as tf
import unittest


class TestReparameterization(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.zero = Expression(('number', 0.0))
        cls.one = Expression(('number', 1.0))
        cls.two = Expression(('+', (cls.one, cls.one)))

        cls.z = Expression(('randomvar', ('Normal', (cls.zero, cls.one))))
        cls.mu = Expression(('pvar_expr', ('mu', ['?x'])))
        cls.sigma = Expression(('pvar_expr', ('sigma', ['?x'])))

        cls.x1 = Expression(('randomvar', ('Normal', (cls.mu, cls.one))))
        cls.x2 = Expression(('randomvar', ('Normal', (cls.zero, cls.sigma))))
        cls.x3 = Expression(('randomvar', ('Normal', (cls.mu, cls.sigma))))

        cls.x4 = Expression(('randomvar', ('Normal', (cls.z, cls.one))))
        cls.x5 = Expression(('randomvar', ('Normal', (cls.x1, cls.one))))
        cls.x6 = Expression(('randomvar', ('Normal', (cls.x1, cls.sigma))))

        cls.mu_plus_z = Expression(('+', (cls.mu, cls.z)))
        cls.z_plus_mu = Expression(('+', (cls.z, cls.mu)))

        cls.mu_plus_x2 = Expression(('+', (cls.mu, cls.x2)))
        cls.x2_plus_mu = Expression(('+', (cls.x2, cls.mu)))

        cls.x1_plus_z = Expression(('+', (cls.x1, cls.z)))
        cls.z_plus_x1 = Expression(('+', (cls.z, cls.x1)))

        cls.z_times_z = Expression(('*', (cls.z, cls.z)))
        cls.x2_times_x2 = Expression(('*', (cls.x2, cls.x2)))

        cls.x7 = Expression(('randomvar', ('Normal', (cls.one, cls.z_times_z))))
        cls.x8 = Expression(('randomvar', ('Normal', (cls.mu, cls.z_times_z))))
        cls.x9 = Expression(('randomvar', ('Normal', (cls.x3, cls.z_times_z))))

        cls.exp_2 = Expression(('func', ('exp', [cls.two])))
        cls.exp_z = Expression(('func', ('exp', [cls.z])))
        cls.exp_x1 = Expression(('func', ('exp', [cls.x1])))

        cls.y1 = Expression(('randomvar', ('Normal', (cls.one, cls.exp_z))))
        cls.y2 = Expression(('randomvar', ('Normal', (cls.mu, cls.exp_z))))
        cls.y3 = Expression(('randomvar', ('Normal', (cls.mu, cls.exp_x1))))

    def test_standard_normal(self):
        noise = get_reparameterization(self.z, scope={})
        self.assertListEqual(noise, [('Normal', [1])])

    def test_multivariate_normal(self):
        scope = {
            'mu/1': TensorFluent(tf.zeros(32), scope=['?x']),
            'sigma/1': TensorFluent(tf.ones(32), scope=['?x'])
        }

        noise1 = get_reparameterization(self.x1, scope=scope)
        self.assertListEqual(noise1, [('Normal', [32])])

        noise2 = get_reparameterization(self.x2, scope=scope)
        self.assertListEqual(noise2, [('Normal', [32])])

        noise3 = get_reparameterization(self.x3, scope=scope)
        self.assertListEqual(noise3, [('Normal', [32])])

        noise4 = get_reparameterization(self.x4, scope=scope)
        self.assertListEqual(noise4, [('Normal', [1]), ('Normal', [1])])

        noise5 = get_reparameterization(self.x5, scope=scope)
        self.assertListEqual(noise5, [('Normal', [32]), ('Normal', [32])])

        noise6 = get_reparameterization(self.x6, scope=scope)
        self.assertListEqual(noise6, [('Normal', [32]), ('Normal', [32])])

        noise7 = get_reparameterization(self.x7, scope=scope)
        self.assertListEqual(noise7, [('Normal', [1]), ('Normal', [1]), ('Normal', [1])])

        noise8 = get_reparameterization(self.x8, scope=scope)
        self.assertListEqual(noise8, [('Normal', [1]), ('Normal', [1]), ('Normal', [32])])

        noise9 = get_reparameterization(self.x9, scope=scope)
        self.assertListEqual(noise9, [('Normal', [32]), ('Normal', [1]), ('Normal', [1]), ('Normal', [32])])

    def test_batch_normal(self):
        scope = {
            'mu/1': TensorFluent(tf.zeros((64, 16)), scope=['?x'], batch=True),
            'sigma/1': TensorFluent(tf.ones((64, 16)), scope=['?x'], batch=True)
        }

        noise1 = get_reparameterization(self.x1, scope=scope)
        self.assertListEqual(noise1, [('Normal', [64, 16])])

        noise2 = get_reparameterization(self.x2, scope=scope)
        self.assertListEqual(noise2, [('Normal', [64, 16])])

        noise3 = get_reparameterization(self.x3, scope=scope)
        self.assertListEqual(noise3, [('Normal', [64, 16])])

    def test_arithmetic(self):
        scope = {
            'mu/1': TensorFluent(tf.zeros(32), scope=['?x']),
            'sigma/1': TensorFluent(tf.ones(32), scope=['?x'])
        }

        noise = get_reparameterization(self.two, scope={})
        self.assertListEqual(noise, [])

        noise = get_reparameterization(self.z_times_z, scope={})
        self.assertListEqual(noise, [('Normal', [1]), ('Normal', [1])])

        noise = get_reparameterization(self.x2_times_x2, scope=scope)
        self.assertListEqual(noise, [('Normal', [32]), ('Normal', [32])])

        noise = get_reparameterization(self.mu_plus_z, scope=scope)
        self.assertListEqual(noise, [('Normal', [1])])

        noise = get_reparameterization(self.z_plus_mu, scope=scope)
        self.assertListEqual(noise, [('Normal', [1])])

        noise = get_reparameterization(self.mu_plus_x2, scope=scope)
        self.assertListEqual(noise, [('Normal', [32])])

        noise = get_reparameterization(self.x2_plus_mu, scope=scope)
        self.assertListEqual(noise, [('Normal', [32])])

        noise = get_reparameterization(self.x1_plus_z, scope=scope)
        self.assertListEqual(noise, [('Normal', [32]), ('Normal', [1])])

        noise = get_reparameterization(self.z_plus_x1, scope=scope)
        self.assertListEqual(noise, [('Normal', [1]), ('Normal', [32])])

    def test_function(self):
        scope = {
            'mu/1': TensorFluent(tf.zeros(24), scope=['?x']),
            'sigma/1': TensorFluent(tf.ones(24), scope=['?x'])
        }

        noise = get_reparameterization(self.exp_2, scope=scope)
        self.assertListEqual(noise, [])

        noise = get_reparameterization(self.exp_z, scope=scope)
        self.assertListEqual(noise, [('Normal', [1])])

        noise = get_reparameterization(self.exp_x1, scope=scope)
        self.assertListEqual(noise, [('Normal', [24])])

        noise = get_reparameterization(self.y1, scope=scope)
        self.assertListEqual(noise, [('Normal', [1]), ('Normal', [1])])

        noise = get_reparameterization(self.y2, scope=scope)
        self.assertListEqual(noise, [('Normal', [1]), ('Normal', [24])])

        noise = get_reparameterization(self.y3, scope=scope)
        self.assertListEqual(noise, [('Normal', [24]), ('Normal', [24])])
